 /* 
  * UAE - The Un*x Amiga Emulator
  * 
  * Support for Alsa sound
  * 
  * Copyright 1997 Bernd Schmidt
  * Copyright (C) 2001 Matthieu Sozeau <mattam@netcourrier.com>
  */

#include "sysconfig.h"
#include "sysdeps.h"

#include "config.h"
#include "options.h"
#include "memory.h"
#include "custom.h"
#include "gensound.h"
#include "sd-sound.h"

#include <sys/ioctl.h>

#include <alsa/asoundlib.h>
#include <alsa/pcm_plugin.h>

#include "uade.h"

static snd_pcm_t *alsa_pcm = 0;
static snd_pcm_t *alsa_pcm_slave = 0;
static snd_pcm_hw_params_t *hwparams = 0;
static snd_pcm_sw_params_t *swparams = 0;
static snd_pcm_format_t alsa_format;
static snd_pcm_status_t *alsa_status = 0;

static unsigned int alsa_rate;
static unsigned int alsa_channels;
static int alsa_buffer_size;
static int alsa_buffer_time;
static int alsa_bits_per_sample;
static int alsa_frame_size;
static int written_frames;
static int alsa_bps = 0;
static long int alsa_total_written = 0;
static char device[7] = "hw:0,0";

static snd_mixer_t *mixer = 0;
static snd_mixer_elem_t *pcm_element = 0;
static snd_mixer_selem_id_t *pcm_elt = 0;

static long int alsa_min_vol, alsa_max_vol;

static int have_sound = 0;

uae_u16 sndbuffer[44100];
uae_u16 *sndbufpt;
int sndbufsize;

void close_sound (void)
{
  if (have_sound) {
    /* close alsa sound */
  }
}

static void alsa_set_volume(int l, int r)
{
  if (!pcm_element)
    goto err_out;
  
  if (snd_mixer_selem_has_playback_volume(pcm_element)) {
    if (snd_mixer_selem_set_playback_volume(pcm_element, SND_MIXER_SCHN_FRONT_LEFT, l))
      goto err_out;
    if (snd_mixer_selem_set_playback_volume(pcm_element, SND_MIXER_SCHN_FRONT_RIGHT, r))
      goto err_out;
  } else {
    fprintf(stderr, "alsa_set_volume: Wrong PCM channel!");
    goto err_out;
  }
  return;

 err_out:
  fprintf(stderr, "alsa_set_volume: error\n");
}

static int alsa_setup_mixer(void)
{
  long int a, b;

  /* alsa mixer setup */
  if (snd_mixer_open(&mixer, 0))
    goto err_out;
  if (snd_mixer_attach(mixer, card_id))
    goto err_out;
  if (snd_mixer_selem_register(mixer, 0, 0))
    goto err_out;
  if (snd_mixer_load (mixer))
    goto err_out;
  if (snd_mixer_selem_id_malloc(&pcm_elt))
    goto err_out;
  
  /* use master */
  pcm_element = snd_mixer_first_elem(mixer);  
  snd_mixer_selem_get_id(pcm_element, pcm_elt);
  while (strcmp(snd_mixer_selem_id_get_name(pcm_elt), "Master")) {
    pcm_element = snd_mixer_elem_next(pcm_element);
    if (!pcm_element)
      break;
    snd_mixer_selem_get_id(pcm_element, pcm_elt);
  }
  
  if (!pcm_element)
    goto err_out;

  snd_mixer_selem_get_playback_volume_range(pcm_element, &alsa_min_vol, &alsa_max_vol);
  snd_mixer_selem_set_playback_volume_range(pcm_element, 0, 100);
  
  snd_mixer_selem_get_playback_volume(pcm_element, SND_MIXER_SCHN_FRONT_LEFT, &a);
  snd_mixer_selem_get_playback_volume(pcm_element, SND_MIXER_SCHN_FRONT_RIGHT, &b);
  alsa_set_volume(a * 100 / alsa_max_vol, b * 100 / alsa_max_vol);

  return 1;

 err_out:
  fprintf(stderr, "alsa_setup_mixer: couldn't setup alsa sound\n");
  return 0;
}


/* Try to determine whether sound is available.  This is only for GUI purposes.  */
int setup_sound (void)
{
  if (uade_local_sound) {
    if (!alsa_setup_mixer()) {
      goto err_out;
    }
  }

  sound_available = 1;
  return 1;

 err_out:
  fprintf(stderr, "setup_sound: couldn't setup alsa sound\n");
  sound_available = 1;
  return 0;
}


static int alsa_setup(int bits, int rate, int channels)
{
  if (bits == 16) {
    alsa_format = SND_PCM_FORMAT_S16_LE;
    fprintf (stderr, "remember to support big endian too!!\n");

  } else if (bits == 8) {
    alsa_format = SND_PCM_FORMAT_S8;
  } else {
    fprintf(stderr, "alsa_setup: illegal number of bits\n");
    return 0;
  }

  alsa_pcm = 0;
  alsa_rate = rate;
  alsa_channels = channels;

  if ((err = snd_pcm_open(&alsa_pcm, "", SND_PCM_STREAM_PLAYBACK, SND_PCM_NONBLOCK)) < 0) {
    fprintf (stderr, "alsa_setup: error opening alsa device\n");
    return 0;
  }

  snd_pcm_hw_params_malloc(&hwparams);
  if(!hwparams) {
    fprintf(stderr, "alsa_setup: no hw params allocated\n");
    return 0;
  }

  if ((err = snd_pcm_hw_params_any(alsa_pcm, hwparams)) < 0) {
    fprintf(stderr, "alsa_setup: broken configuration for playback: no configuration available\n");
    return 0;
  }

  snd_pcm_hw_params_set_access(alsa_pcm, hwparams, SND_PCM_ACCESS_RW_INTERLEAVED);
  if ((err = snd_pcm_hw_params_set_format(alsa_pcm, hwparams, alsa_format)) < 0) {
    fprintf(stderr, "alsa_setup: sample format not available for playback\n");
    return 0;
  }

  snd_pcm_hw_params_set_channels(alsa_pcm, hwparams, alsa_channels);

  chosen_rate = snd_pcm_hw_params_set_rate_near(alsa_pcm, hwparams, alsa_rate, 0);
  if (chosen_rate < 0) {
    fprintf(stderr, "alsa_setup: can't set rate\n");
    return 0;
  }

  if (chosen_rate != alsa_rate) {
    /* oh hell.. we would have to deal with this. perhaps in a later version */
    fprintf(stderr, "alsa_setup: chosen_rate differs (%d, %d)\n", chosen_rate, alsa_rate);
  }

  alsa_buffer_time = snd_pcm_hw_params_set_buffer_time_near(alsa_pcm, hwparams, 500 * 1000, 0);
  if (alsa_buffer_time < 0) {
    fprintf(stderr, "alsa_setup: buffer_time < 0\n");
    return 0;
  }
  
  if (snd_pcm_hw_params_set_period_time_near(alsa_pcm, hwparams, alsa_buffer_time / 4, 0) < 0) {
    fprintf(stderr, "alsa_setup: period_time < 0\n");
    return 0;
  }
  
  snd_pcm_hw_params(alsa_pcm, hwparams);

  alsa_buffer_size = snd_pcm_hw_params_get_buffer_size(hwparams);
  if (snd_pcm_hw_params_get_period_size(hwparams, 0) == alsa_buffer_size) {
    fprintf (stderr, "alsa_setup: period_size doesn't match alsa_buffer size\n");
    return 0;
  }
  
  swparams = (snd_pcm_sw_params_t *) malloc(snd_pcm_sw_params_sizeof());
  if (!swpamars) {
    fprintf(stderr, "alsa_setup: no memory for swparams\n");
    return 0;
  }

  snd_pcm_sw_params_current(alsa_pcm, swparams);
  if (!swparams) {
    fprintf (stderr, "alsa_setup: no swparams\n");
    return 0;
  }
  
  if (snd_pcm_sw_params(alsa_pcm, swparams) < 0) {
    fprintf(stderr, "alsa_setup: unable to install swparams\n");
    return 0;
  }
  
  alsa_bits_per_sample = snd_pcm_format_physical_width(alsa_format); 
  alsa_frame_size = (alsa_bits_per_sample * alsa_channels) / 8;
  alsa_bps = (alsa_rate * alsa_frame_size) / 1000;
  return 1;
}

int init_sound (void)
{
  int stereo;
  int rate;
  int dspbits;

  if (currprefs.sound_maxbsiz < 128 || currprefs.sound_maxbsiz > 16384) {
    fprintf (stderr, "Sound buffer size %d out of range.\n", currprefs.sound_maxbsiz);
    currprefs.sound_maxbsiz = 8192;
  }

  dspbits = currprefs.sound_bits;
  rate    = currprefs.sound_freq;
  stereo  = currprefs.stereo;
  have_sound = 1;
  sndbufsize = 4096;

  if (uade_local_sound) {
    /* alsa open sound */
    if (alsa_pcm)
      alsa_close();

    alsa_setup(dspbits, rate, stereo ? 2 : 1);
    alsa_setup_mixer();

    snd_pcm_prepare(alsa_pcm);  
  }

  sample_evtime = (long) maxhpos * maxvpos * 50 / rate;
  if (dspbits == 16) {
    init_sound_table16 ();
    sample_handler = currprefs.stereo ? sample16s_handler : sample16_handler;
  } else {
    init_sound_table8 ();
    sample_handler = currprefs.stereo ? sample8s_handler : sample8_handler;
  }
  sound_available = 1;

  sndbufpt = sndbuffer;
#ifdef FRAME_RATE_HACK
  vsynctime = vsynctime * 9 / 10;
#endif	
  return 1;
}

/* this should be called between subsongs when remote slave changes subsong */
void flush_sound (void)
{
  sndbufpt = sndbuffer;
}
